import re
import typing
import uuid

import marshmallow
import requests_mock
from requests_mock import create_response

from commercetools import schemas, types
from commercetools.client import CommercetoolsError
from commercetools.schemas import ResourceSchema
from commercetools.services import abstract
from commercetools.testing import utils
from commercetools.testing.predicates import PredicateFilter


class BaseModel:
    _resource_schema: typing.Type[ResourceSchema]

    def __init__(self, storage):
        self._storage = storage
        self.objects = storage.init_store(self._primary_type_name)
        assert (
            self._resource_schema
        ), f"{self.__class__.__name__} has no resource schema defined"

    def add(self, obj, id=None):
        obj = self._create_from_draft(obj, id)
        key = self._generate_key(obj)
        document = self._resource_schema().dump(obj)
        self.objects[key] = document
        return document

    def _generate_key(self, obj):
        return uuid.UUID(obj.id)

    def _create_from_draft(self, obj, id=None):
        raise NotImplementedError()

    def query(self, where):
        objects = list(self.objects.values())
        if not where:
            return objects
        where_clause = " AND ".join(where)
        predicate_filter = PredicateFilter(where_clause)
        return [obj for obj in objects if predicate_filter.match(obj)]

    def get_by_id(self, id):
        try:
            return self.objects.get(uuid.UUID(id))
        except ValueError:
            return None

    def get_by_key(self, key):
        for obj in self.objects.values():
            if obj["key"] == key:
                return obj

    def delete_by_id(self, id):
        obj = self.objects.pop(uuid.UUID(id))
        return obj

    def delete_by_key(self, key):
        for obj_id, obj in self.objects.items():
            if obj["key"] == key:
                return self.objects.pop(obj_id)

    def save(self, obj):
        assert obj["id"]
        obj["version"] += 1
        key = uuid.UUID(obj["id"])
        self.objects[key] = obj


class BaseBackend:
    path = None
    _actions: typing.Dict[str, typing.Callable] = {}

    def __init__(self, storage=None, model=None):
        if model:
            self.model = model
        elif storage:
            self.model = self.model_class(storage)
        else:
            self.model = self.model_class()

    def register(self, adapter):
        adapter.add_matcher(self._matcher)

    def _matcher(self, request, skip_port_check=False):
        if not skip_port_check and request.port != 443:
            return

        if request.hostname not in self.hostnames:
            return
            if not request.route_kwargs["path"]:
                request.route_kwargs["path"] = "/"

        match = re.match(self.path_prefix, request.path)
        if match:
            request.kwargs = match.groupdict()
            try:
                request_path = match.groupdict()["path"]
            except KeyError:
                request_path = ""

            for path, method, callback in self.urls():
                path_match = re.match(path, request_path)

                # Call the view
                if path_match and request.method == method:
                    response = callback(request, **path_match.groupdict())
                    if response is None:
                        raise NotImplementedError(
                            "No response returned by %r" % callback
                        )
                    return response
            return requests_mock.create_response(request, status_code=404)


class ServiceBackend(BaseBackend):
    hostnames = ["api.sphere.io", "localhost"]
    model_class: typing.Any = None
    _schema_draft: typing.Optional[marshmallow.Schema] = None
    _schema_update: typing.Optional[marshmallow.Schema] = None
    _schema_query_response: typing.Optional[marshmallow.Schema] = None

    @property
    def path_prefix(self):
        return f"/(?P<project>[^/]+)/{self.service_path}/?(?P<path>.*)?"

    def create(self, request):
        obj = self._schema_draft().loads(request.body)
        data = self.model.add(obj)
        return create_response(request, json=data)

    def get_by_id(self, request, id):
        obj = self.model.get_by_id(id)
        if obj:
            return create_response(request, json=obj)
        return create_response(request, status_code=404)

    def get_by_key(self, request, key):
        obj = self.model.get_by_key(key)
        if obj:
            return create_response(request, json=obj)
        return create_response(request, status_code=404)

    def query(self, request):
        params = utils.parse_request_params(abstract.AbstractQuerySchema, request)
        results = self.model.query(params.get("where"))
        total_count = len(results)
        if params.get("limit"):
            results = results[: params["limit"]]

        data = {
            "count": len(results),
            "total": total_count,
            "offset": 0,
            "results": self.model._resource_schema().load(results, many=True),
        }
        content = self._schema_query_response().dumps(data)
        return create_response(request, text=content)

    def update_by_id(self, request, id):
        obj = self.model.get_by_id(id)
        return self._update(request, obj)

    def update_by_key(self, request, key):
        obj = self.model.get_by_key(key)
        return self._update(request, obj)

    def delete_by_id(self, request, id):
        obj = self.model.get_by_id(id)
        if obj:
            response = self._validate_resource_version(request, obj)
            if response is not None:
                return response

            obj = self.model.delete_by_id(id)
            return create_response(request, json=obj)
        return create_response(request, status_code=404)

    def delete_by_key(self, request, key):
        obj = self.model.get_by_key(key)
        if obj:
            response = self._validate_resource_version(request, obj)
            if response is not None:
                return response

            obj = self.model.delete_by_key(key)
            return create_response(request, json=obj)
        return create_response(request, status_code=404)

    def _update(self, request, obj):
        if not obj:
            return create_response(request, status_code=404)

        update = self._schema_update().load(request.json())
        if update.actions:
            obj, err = self._apply_update_actions(obj, update)
            if err:
                return create_response(
                    request, json=err, status_code=err["statusCode"]
                )
        return create_response(request, json=obj)

    def _validate_resource_version(self, request, obj):
        update_version = self._get_version_from_request(request)
        if update_version != obj["version"]:
            data = self._create_error_response(obj["version"])
            return create_response(request, json=data, status_code=409)

    def _get_version_from_request(self, request):
        version_data = request.qs.get("version")
        if version_data:
            return int(version_data[0])
        return request.json().get("version")

    def _apply_update_actions(self, obj, update):
        original_obj = obj

        for action in update.actions:
            func = self._actions.get(action.action)
            if not func:
                print("Missing action for", action.action)
                continue
            obj = func(self, obj, action)

        # Save the updated object to the model
        if obj != original_obj:
            if obj["version"] != update.version:
                return None, self._create_error_response(obj["version"])
            self.model.save(obj)

        # Temporary
        elif not self._actions:
            self.model.save(obj)

        return obj, None

    def _create_error_response(self, version):
        return schemas.ErrorResponseSchema().dump(
            types.ErrorResponse(
                status_code=409,
                message="Version mismatch. Concurrent modification.",
                errors=[
                    types.ConcurrentModificationError(
                        message="Version mismatch. Concurrent modification.",
                        current_version=version,
                    )
                ],
            )
        )
